# walks the program from a serie of entrypoints, returns a list of block -> block in a function and function -> function (2 hashes)
def dump_makegraph(dasm, eplist)
	# funcaddr => [list of subfuncaddr]
	func_rel = {}
	# blockaddr => [list of blockaddr] (same func only)
	blk_rel = {}

	# build graph
	todo_f = eplist.dup
	done_f = []
	while f = todo_f.shift
		f = dasm.normalize f
		next if done_f.include? f
		done_f << f
		func_rel[f] = []

		todo_b = [f]
		done_b = []
		while b = todo_b.shift
			next if done_b.include? b
			done_b << b
			blk_rel[b] = []
			next if not di = dasm.decoded[b]
			di.block.each_to_indirect { |t|
				t = dasm.normalize t
				todo_f << t
				func_rel[f] << t
			}
			if di.block.to_subfuncret
				di.block.each_to_normal { |t|
					t = dasm.normalize t
					todo_f << t
					func_rel[f] << t
				}
				di.block.each_to_subfuncret { |t|
					t = dasm.normalize t
					todo_b << t
					blk_rel[b] << t
				}
			else
				di.block.each_to_normal { |t|
					t = dasm.normalize t
					todo_b << t
					blk_rel[b] << t
				}
			end
			blk_rel[b].uniq!
		end
		func_rel[f].uniq!
	end
	[func_rel, blk_rel]
end


def dump_graphml(filename, dasm, eplist)
	func_rel, blk_rel = dump_makegraph(dasm, eplist)

	# output file
	File.open(filename, 'w') { |fd|
		# stdhdr
		fd.puts '<?xml version="1.0" encoding="UTF-8" standalone="no"?>', '<graphml xmlns="http://graphml.graphdrawing.org/xmlns/graphml" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:y="http://www.yworks.com/xml/graphml" xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns/graphml http://www.yworks.com/xml/schema/graphml/1.0/ygraphml.xsd">', '<key for="node" id="d0" yfiles.type="nodegraphics"/>', '<key for="edge" id="d2" yfiles.type="edgegraphics"/>'

		# toplevel graph: functions
		fd.puts '<graph edgedefault="directed" id="funcs">'

		todo_f = eplist.dup
		done_f = []
		done_b = []
		while f = todo_f.shift
			next if done_f.include? f
			done_f << f
			
			fd.puts '<node id="f_' << Expression[f].to_s << '">', '<data key="d0"><y:ShapeNode><y:NodeLabel>' << (dasm.auto_label_at(f, 'func') || Expression[f].to_s) << '</y:NodeLabel></y:ShapeNode></data>'
			if blk_rel[f]
				# nested graph: function body
				fd.puts ' <graph edgedefault="directed" id="func_' << Expression[f].to_s << '">'
				todo_b = [f]
				while b = todo_b.shift
					next if done_b.include? b
					done_b << b

					text = dump_graphml_block(dasm, b)
					text = '<y:Geometry height="' << (16*text.to_a.length).to_s << '" width="' << (text.empty? ? '10' : '450') << '"/>' << 
					'<y:NodeLabel alignment="left" autoSizePolicy="node_height" fontFamily="Monospaced" fontSize="12" fontStyle="plain" hasBackgroundColor="false" hasLineColor="false" height="128.0" modelName="internal" modelPosition="l" textColor="#000000" visible="true" width="455.0" x="4.0" y="0.0">' << text << '</y:NodeLabel>'
					fd.puts ' <node id="b_' << Expression[b].to_s << '"><data key="d0"><y:ShapeNode>' << text << '</y:ShapeNode></data></node>'

					blk_rel[b].to_a.each { |tb|
						fd.puts ' <edge source="b_' << Expression[b].to_s << '" target="b_' << Expression[tb].to_s << '"><data key="d2"><y:PolyLineEdge><y:Arrows source="none" target="standard"/></y:PolyLineEdge></data></edge>'
						todo_b << tb
					}
				end
				fd.puts ' </graph>'
			end
			fd.puts '</node>'

			func_rel[f].to_a.each { |tf|
				fd.puts '<edge source="f_' << Expression[f].to_s << '" target="f_' << Expression[tf].to_s << '"><data key="d2"><y:PolyLineEdge><y:Arrows source="none" target="standard"/></y:PolyLineEdge></data></edge>'
				todo_f << tf
			}
		end
		fd.puts '</graph>', '</graphml>'
	}
end

def dump_graphml_block(dasm, addr)
	return Expression[addr].to_s if not di = dasm.decoded[addr]
	ret = ''
	list = di.block.list.find_all { |di| di.instruction.opname != 'nop' or (di.opcode.name == 'call' and di.block.to_subfuncret) }
	dasm.dump_block_header(dasm.decoded[addr].block) { |l| ret << l << "\n" } if not list.empty?
	list.each { |di| ret << di.show << "\n" }
	ret.gsub('&', '&amp;').gsub('<', '&lt;').gsub('>', '&gt;').gsub(/\A(?:\s*\n)+|(?:\s*\n)+\Z/, '')
end

def dump_asm(filename, dasm, eplist)
	func_rel, blk_rel = dump_makegraph(dasm, eplist)

	# output file
	File.open(filename, 'w') { |fd|
		todo_f = eplist.dup
		done_f = []
		done_b = []
		while f = todo_f.shift
			next if done_f.include? f
			done_f << f
			
			if blk_rel[f]
				todo_b = [f]
				while b = todo_b.pop
					next if done_b.include? b or not dasm.decoded[b]
					done_b << b

					block = dasm.decoded[b].block
					dasm.dump_block_header(block) { |l| fd.puts l }
					block.list.each { |di|
						fd.puts di.show if di.instruction.opname != 'nop'
						if di.opcode.name == 'call' and di.comment.to_a.include? 'noreturn'
							retaddr = di.address + di.bin_length
							if dasm.decoded[retaddr]
								fd.puts "    jmp #{dasm.auto_label_at retaddr}    ; blabla ret"
							else
								fd.puts "    hlt    ; blabla XXX"
							end
							fd.puts
							fd.puts "#{di.instruction.args.first}:" if di != block.list.last
						end
					}

					to = blk_rel[b].to_a
					todo_b.concat to
					op = block.list.last.opcode if block.list.last
					if to.length == 2 and not op.props[:stopexec] and op.name[0] == ?j and not block.list.empty?
						following = block.list.last.address + block.list.last.bin_length
						todo_b << following
						to = [following]
					end
					case to.length
					when 0
						fd.puts
					when 1
						if done_b.include? to.first
							fd.puts "    jmp #{dasm.auto_label_at to.first}	; blabla"
							fd.puts
						end
					else
						fd.puts "; blabla XXX to: #{to.map { |t| Expression[t] }.join(', ')}"
						fd.puts
					end
				end
			end

			func_rel[f].to_a.each { |tf| todo_f << tf }
		end
		fd.puts
	}
end
