require 'metasm'
include Metasm

file = 'poeut.exe'
puts 'reading original'
pe = PE.decode_file 'poeut.exe'
pe.cpu = pe.cpu_from_headers
elf = ELF.new pe.cpu
elf.header.ident = 0.chr*14+4.chr+0.chr		# chpax -m

puts 'copy data/bss sections'
pe.sections.each { |s|
	es = ELF::Section.new
	es.name = ".orig#{pe.sections.index(s)}"
	es.type = 'PROGBITS'
	es.flags = ['ALLOC', 'WRITE']
	es.addr = s.virtaddr + pe.optheader.image_base
	es.encoded = s.encoded
	elf.sections << es
}

puts 'parse asm'
src = File.read('poeut.asm').sub('entrypoint:', '')
elf.new_label 'entrypoint'
elf.parse DATA.read
elf.parse src


pe.parse '.section ".clrtext" rx'
pe.parse '.entrypoint'
pe.parse src

puts 'assemble'
elf.assemble
pe.assemble

puts 'fixup data relocs'
text = pe.sections.last.encoded
text.fixup! 'loc_0' => 0, 'loc_1' => 1
text.reloc.values.map { |r| r.target.reduce_rec }.grep(::String).uniq.sort.each { |t|
	if t =~ /(?:dword|byte)_(4\w+)h/ and not text.export[t]
		rva = $1.to_i(16) - pe.optheader.image_base
		if not s = pe.sections.find { |s| s.virtaddr <= rva and s.virtaddr + s.virtsize >= rva }
			raise "#{Expression[rva]} #{t}"
		else
			s.encoded.add_export t, rva-s.virtaddr
		end
	end
}

elf.encode_file 'unpoeut'
pe.encode_file 'unpoeut.exe'

__END__
.pt_gnu_stack rw
.text
.entrypoint
	; need a huuuuge stack => heap
	push 0x80100
	call malloc
	add esp, 4
	lea ebp, [eax+0x80000]

	; disable gnulibc init (weird things w/ FindAtomA)
	or dword ptr [40fc00h], 1

	; init argc/argv/envp as args for main
	mov ecx, [esp]
	mov [ebp], ecx
	lea eax, [esp+4]
	mov [ebp+4], eax
	lea eax, [eax+4*ecx+4]
	mov [ebp+8], eax

	; call main
	mov esp, ebp
	call loc_403db0h
	
	; exit
	mov eax, 1
	mov ebx, 0
	int 80h

puts:
	push [esp+4]
	push puts_format
	call printf
	add esp, 8
	ret
puts_format db "%s\n\0"

AddAtomA:
ExitProcess:
FindAtomA:
GetAtomNameA:
SetUnhandledExceptionFilter:
signal:
_cexit:
_iob:
_onexit:
_setmode:
__getmainargs:
__p__environ:
__p__fmode:
__set_app_type:
loc_0:
loc_1:
	hlt
	jmp iat_atexit

orig_entrypoint:
// concat unobfuscated src here
